from typing import List

from personmatchmodule.models.constants import Attribute
from personmatchmodule.models.rule import Rule


class RulesGenerator:
    @staticmethod
    def generate_rules() -> List[Rule]:
        """
        generate default match rules
        :return: generated rules for matching
        """

        rules: List[Rule] = []

        rules.append(
            Rule(
                name="Rule-001",
                description="given name, family name, gender, dob, zip",
                number=1,
                attributes=[
                    Attribute.NAME_GIVEN,
                    Attribute.NAME_FAMILY,
                    Attribute.GENDER,
                    Attribute.BIRTH_DATE,
                    Attribute.ADDRESS_POSTAL_CODE,
                ],
                score=0.0,
            )
        )
        rules.append(
            Rule(
                name="Rule-002",
                description="given name, dob, address 1, zip",
                number=2,
                attributes=[
                    Attribute.NAME_GIVEN,
                    Attribute.BIRTH_DATE,
                    Attribute.ADDRESS_LINE_1,
                    Attribute.ADDRESS_POSTAL_CODE,
                ],
                score=0.0,
            )
        )
        rules.append(
            Rule(
                name="Rule-003",
                description="given name, date of birth, email",
                number=3,
                attributes=[
                    Attribute.NAME_GIVEN,
                    Attribute.BIRTH_DATE,
                    Attribute.EMAIL,
                ],
                score=0.0,
            )
        )
        rules.append(
            Rule(
                name="Rule-004",
                description="given name, date of birth, phone",
                number=4,
                attributes=[
                    Attribute.NAME_GIVEN,
                    Attribute.BIRTH_DATE,
                    Attribute.PHONE,
                ],
                score=0.0,
            )
        )
        rules.append(
            Rule(
                name="Rule-005",
                description="given name, family name, year of date of birth, gender, address 1, zip",
                number=5,
                attributes=[
                    Attribute.NAME_GIVEN,
                    Attribute.NAME_FAMILY,
                    Attribute.BIRTH_DATE_YEAR,
                    Attribute.GENDER,
                    Attribute.ADDRESS_LINE_1,
                    Attribute.ADDRESS_POSTAL_CODE,
                ],
                score=0.0,
            )
        )
        rules.append(
            Rule(
                name="Rule-006",
                description="given name, family name, dob month, dob date, gender, address 1, zip",
                number=6,
                attributes=[
                    Attribute.NAME_GIVEN,
                    Attribute.NAME_FAMILY,
                    Attribute.BIRTH_DATE_MONTH,
                    Attribute.BIRTH_DATE_DAY,
                    Attribute.GENDER,
                    Attribute.ADDRESS_LINE_1,
                    Attribute.ADDRESS_POSTAL_CODE,
                ],
                score=0.0,
            )
        )
        rules.append(
            Rule(
                name="Rule-007",
                description="given name, family name, date of birth, gender, phone",
                number=7,
                attributes=[
                    Attribute.NAME_GIVEN,
                    Attribute.NAME_FAMILY,
                    Attribute.BIRTH_DATE,
                    Attribute.GENDER,
                    Attribute.PHONE_AREA,
                    Attribute.PHONE_LOCAL,
                ],
                score=0.0,
            )
        )
        rules.append(
            Rule(
                name="Rule-008",
                description="first name, last name, date of birth, gender, phone local exchange, phone line",
                number=8,
                attributes=[
                    Attribute.NAME_GIVEN,
                    Attribute.NAME_FAMILY,
                    Attribute.BIRTH_DATE,
                    Attribute.GENDER,
                    Attribute.PHONE_LOCAL,
                    Attribute.PHONE_LINE,
                ],
                score=0.0,
            )
        )
        rules.append(
            Rule(
                name="Rule-009",
                description="first name, last name, date of birth, gender, phone area code, phone line",
                number=9,
                attributes=[
                    Attribute.NAME_GIVEN,
                    Attribute.NAME_FAMILY,
                    Attribute.BIRTH_DATE,
                    Attribute.GENDER,
                    Attribute.PHONE_AREA,
                    Attribute.PHONE_LINE,
                ],
                score=0.0,
            )
        )
        rules.append(
            Rule(
                name="Rule-10",
                description="given name, dob, gender, address 1 street number, zip, email username, phone line",
                number=10,
                attributes=[
                    Attribute.NAME_GIVEN,
                    Attribute.BIRTH_DATE,
                    Attribute.GENDER,
                    Attribute.ADDRESS_LINE_1_ST_NUM,
                    Attribute.ADDRESS_POSTAL_CODE,
                    Attribute.EMAIL_USERNAME,
                    Attribute.PHONE_LINE,
                ],
                score=0.0,
            )
        )
        rules.append(
            Rule(
                name="Rule-011",
                description="given name, dob, gender, address 1 street number, zip, phone area code, phone local exchange code",
                number=11,
                attributes=[
                    Attribute.NAME_GIVEN,
                    Attribute.BIRTH_DATE,
                    Attribute.GENDER,
                    Attribute.ADDRESS_LINE_1_ST_NUM,
                    Attribute.ADDRESS_POSTAL_CODE,
                    Attribute.PHONE_AREA,
                    Attribute.PHONE_LOCAL,
                ],
                score=0.0,
            )
        )
        rules.append(
            Rule(
                name="Rule-012",
                description="given name, dob, gender, address 1 street number, zip, phone area code, phone line number",
                number=12,
                attributes=[
                    Attribute.NAME_GIVEN,
                    Attribute.BIRTH_DATE,
                    Attribute.GENDER,
                    Attribute.ADDRESS_LINE_1_ST_NUM,
                    Attribute.ADDRESS_POSTAL_CODE,
                    Attribute.PHONE_AREA,
                    Attribute.PHONE_LINE,
                ],
                score=0.0,
            )
        )
        rules.append(
            Rule(
                name="Rule-013",
                description="given name, dob, gender, address 1 street number, zip, phone local exchange code, phone line number",
                number=13,
                attributes=[
                    Attribute.NAME_GIVEN,
                    Attribute.BIRTH_DATE,
                    Attribute.GENDER,
                    Attribute.ADDRESS_LINE_1_ST_NUM,
                    Attribute.ADDRESS_POSTAL_CODE,
                    Attribute.PHONE_LOCAL,
                    Attribute.PHONE_LINE,
                ],
                score=0.0,
            )
        )
        rules.append(
            # adult_today_flag means that person is likely not young living at home with twin with same DOB, family name, address, and phone.
            Rule(
                name="Rule-014",
                description="family name, date of birth, is adult today flag, gender, address 1, zip, phone",
                number=14,
                attributes=[
                    Attribute.NAME_FAMILY,
                    Attribute.BIRTH_DATE,
                    Attribute.IS_ADULT_TODAY,
                    Attribute.GENDER,
                    Attribute.ADDRESS_LINE_1,
                    Attribute.ADDRESS_POSTAL_CODE,
                    Attribute.PHONE,
                ],
                score=0.0,
            )
        )
        rules.append(
            # adult_today_flag means that person is likely not young living at home with twin who has the same DOB, family name, address, and email.
            Rule(
                name="Rule-015",
                description="family name, date of birth, is adult today flag, gender, address 1, zip, email",
                number=15,
                attributes=[
                    Attribute.NAME_FAMILY,
                    Attribute.BIRTH_DATE,
                    Attribute.IS_ADULT_TODAY,
                    Attribute.GENDER,
                    Attribute.ADDRESS_LINE_1,
                    Attribute.ADDRESS_POSTAL_CODE,
                    Attribute.EMAIL,
                ],
                score=0.0,
            )
        )
        rules.append(
            Rule(
                name="Rule-016",
                description="given name, email, phone, dob_year",
                number=16,
                attributes=[
                    Attribute.NAME_GIVEN,
                    Attribute.EMAIL,
                    Attribute.PHONE,
                    Attribute.BIRTH_DATE_YEAR,
                ],
                score=0.0,
            )
        )
        rules.append(
            Rule(
                name="Rule-050",
                description="SSN",
                number=50,
                attributes=[
                    Attribute.SSN,
                ],
                score=0.0,
            )
        )
        rules.append(
            Rule(
                name="Rule-060",
                description="client slug, given_name, dob_year",
                number=60,
                attributes=[
                    Attribute.META_SECURITY_CLIENT_SLUG,
                    Attribute.NAME_GIVEN,
                    Attribute.BIRTH_DATE_YEAR,
                ],
                score=0.0,
            )
        )

        return rules



from typing import List


class Rule:
    def __init__(
        self,
        name: str,
        description: str,
        number: int,
        attributes: List[str],
        score: float,
    ) -> None:
        self.name: str = name
        self.description: str = description
        self.number: int = number
        self.attributes: List[str] = attributes
        self.score: float = score



class Attribute:
    # Rule
    RULE_ID: str = "rule_id"
    RULE_NAME: str = "rule_name"
    RULE_DESC: str = "rule_desc"
    RULE_SCORE: str = "rule_score"

    # FHIR Person or Patient attributes
    ID: str = "id"
    META_SECURITY_CLIENT_SLUG: str = "meta_security_client_slug"
    NAME_USE: str = "name_use"  # required, "usual | official | temp | nickname | anonymous | old | maiden"
    NAME_TEXT: str = "name_text"
    NAME_GIVEN: str = "name_given"
    NAME_FAMILY: str = "name_family"
    TELECOM_SYSTEM: str = (
        "telecom_system"  # required, "phone | fax | email | pager | url | sms | other"
    )
    TELECOM_USE: str = "telecom_use"  # required, "home | work | temp | old | mobile"
    TELECOM_VALUE: str = "telecom_value"
    GENDER: str = "gender"
    BIRTH_DATE: str = "birth_date"
    ADDRESS_USE: str = "address_use"  # required, "home | work | temp | old | billing"
    ADDRESS_TYPE: str = "address_type"  # required, "postal | physical | both"
    ADDRESS_LINE_1: str = "address_line_1"
    ADDRESS_LINE_2: str = "address_line_2"
    ADDRESS_CITY: str = "address_city"
    ADDRESS_STATE: str = "address_state"
    ADDRESS_POSTAL_CODE: str = "address_postal_code"
    ADDRESS_COUNTRY: str = "address_country"
    ACTIVE: str = "active"
    LINK_TARGET: str = "link_target"
    LINK_ASSURANCE: str = (
        "link_assurance"  # required, "level1 | level2 | level3 | level4"
    )

    # other attributes specifics, needs to be parsed
    EMAIL: str = "email"  # WHEN telecom.system = "email"
    EMAIL_USERNAME: str = "email_username"  # email username, string before the "@"

    PHONE: str = "phone"  # WHEN telecom.system = "phone"
    PHONE_AREA: str = "phone_area"  # phone area number (first 3 digit)
    PHONE_LOCAL: str = "phone_local"  # phone local number (middle 3 digit)
    PHONE_LINE: str = "phone_line"  # phone line number (last 4 digit)

    BIRTH_DATE_YEAR: str = "birth_date_year"  # YYYY from birthDate
    BIRTH_DATE_MONTH: str = "birth_date_month"  # MM from birthDate
    BIRTH_DATE_DAY: str = "birth_date_day"  # DD from birthDate

    ADDRESS_LINE_1_ST_NUM: str = "address_line_1_st_num"

    # this is a "calculate-on-the-fly" attribute,
    #   IS_ADULT_TODAY = (today's Date - birthDate) >= 18 years old
    IS_ADULT_TODAY: str = "is_adult_today"

    # ssn can be under the identifier array, identifier.value
    #   WHEN identifier.system = "http://hl7.org/fhir/sid/us-ssn"
    SSN: str = "ssn"

    # Scores
    SCORE: str = "score"
    TOTAL_SCORE: str = "total_score"
    MAX_POSSIBLE_SCORE: str = "max_possible_score"



from typing import Any, Dict, List

from personmatchmodule.logics.rules_generator import RulesGenerator
from personmatchmodule.logics.score_calculator import ScoreCalculator
from personmatchmodule.models.constants import Attribute
from personmatchmodule.models.rule import Rule


def test_person_one_to_one_match() -> None:
    person_1: Dict[str, Any] = {
        Attribute.ID: "unitypoint-b3a42xc8",
        Attribute.NAME_FAMILY: "bourne",
        Attribute.NAME_GIVEN: "James ",
        Attribute.GENDER: "MALE",
        Attribute.BIRTH_DATE: "1980-09-17",
        Attribute.BIRTH_DATE_YEAR: "1980",  # parsed from birth date
        Attribute.BIRTH_DATE_MONTH: "09",  # parsed from birth date
        Attribute.BIRTH_DATE_DAY: "17",  # parsed from birth date
        Attribute.ADDRESS_LINE_1: "363 Mackerline St",
        Attribute.ADDRESS_LINE_1_ST_NUM: "363",  # parsed from address
        Attribute.ADDRESS_POSTAL_CODE: "34239",
        Attribute.EMAIL: "jbourne@gmail.com",
        Attribute.EMAIL_USERNAME: "jborune",  # parsed from the email
        Attribute.PHONE: "7770003333",
        Attribute.PHONE_AREA: "777",  # parsed from phone
        Attribute.PHONE_LOCAL: "000",  # parsed from phone
        Attribute.PHONE_LINE: "3333",  # parsed from phone
        Attribute.IS_ADULT_TODAY: "true",  # calculated by today's date - birth date >= 18 yrs old
        Attribute.SSN: "",  # parsed from identifier.value
        Attribute.META_SECURITY_CLIENT_SLUG: "unitypoint",  # parsed from meta.security access
    }

    person_2: Dict[str, Any] = {
        Attribute.ID: "unitypoint-73bvy14ka0",
        Attribute.NAME_FAMILY: "BORNE",
        Attribute.NAME_GIVEN: " JAMES ",
        Attribute.GENDER: "male",
        Attribute.BIRTH_DATE: "1980-09-17",
        Attribute.BIRTH_DATE_YEAR: "1980",  # parsed from birth date
        Attribute.BIRTH_DATE_MONTH: "09",  # parsed from birth date
        Attribute.BIRTH_DATE_DAY: "17",  # parsed from birth date
        Attribute.ADDRESS_LINE_1: "363 mackerline st.",
        Attribute.ADDRESS_LINE_1_ST_NUM: "363",  # parsed from address
        Attribute.ADDRESS_POSTAL_CODE: "34239",
        Attribute.EMAIL: "jbourne@hotmail.com",
        Attribute.EMAIL_USERNAME: "jborune",  # parsed from the email
        Attribute.PHONE: "7770003355",
        Attribute.PHONE_AREA: "777",  # parsed from phone
        Attribute.PHONE_LOCAL: "000",  # parsed from phone
        Attribute.PHONE_LINE: "3355",  # parsed from phone
        Attribute.IS_ADULT_TODAY: "true",  # calculated by today's date - birth date >= 18 yrs old
        Attribute.SSN: "555992222",
        Attribute.META_SECURITY_CLIENT_SLUG: "unitypoint",
    }

    rules: List[Rule] = RulesGenerator.generate_rules()

    # match on first rule
    rule_score_result = ScoreCalculator.calculate_score_for_rule(
        rules[0], person_1, person_2
    )
    assert (
        rule_score_result is not None
        and rule_score_result.get(Attribute.ID + "_1") is not None
        and rule_score_result.get(Attribute.RULE_SCORE) is not None
    )
    print(rule_score_result)

    # match on all rules
    rules_score_results: List[Dict[str, Any]] = ScoreCalculator.calculate_score(
        rules, person_1, person_2
    )
    assert len(rules_score_results) == len(rules)
    print(rules_score_results)

    # calculate the final score
    final_score: float = 0.0
    for result in rules_score_results:
        final_score += float(str(result.get(Attribute.RULE_SCORE)))
    final_score /= len(rules)
    print(f"FINAL SCORE: {final_score}")

    # SUGGESTION
    # based on the calculated final score, in 0-100 scale,
    #  the caller categorize and assign the FHIR AssuranceLevel when creating data linkage,
    #  https://www.hl7.org/fhir/valueset-identity-assuranceLevel.html#expansion
    #  "level1" (the lowest): 0.0 - 25.0 for little or no confidence in the asserted identity's accuracy
    #  "level2"             : 25.0 - 50.0 for some confidence
    #  "level3"             : 50.0 - 80.0 for high confidence
    #  "level4" (the highest): 80.0 - 100.0 for very high confidence



from typing import Optional, List, Dict, Any

from rapidfuzz import fuzz

from personmatchmodule.models.constants import Attribute
from personmatchmodule.models.rule import Rule


class ScoreCalculator:
    @staticmethod
    def initialize_score(rules: List[Rule]) -> None:
        for rule in rules:
            rule.score = 0.0

    @staticmethod
    def calculate_score(
        rules: List[Rule], data_1: Dict[str, Any], data_2: Dict[str, Any]
    ) -> List[Dict[str, Any]]:
        """
        Calculate matching scores for ALL rules between FHIR Person-Person, or Person-Patient, or Person/Patient-AppUser
        :param rules: generated rules by RulesGenerator
        :param data_1: Dictionary of Pii data for FHIR Person/Patient data, or AppUser data
        :param data_2: Dictionary of Pii data for FHIR Person/Patient data, or AppUser data
        :return: list of dictionary for rules score results for all rules
        """

        rules_score_results: List[Dict[str, Any]] = []

        for rule in rules:
            rule_score_result = ScoreCalculator.calculate_score_for_rule(
                rule, data_1, data_2
            )
            if rule_score_result is not None:
                rules_score_results.append(rule_score_result)

        return rules_score_results

    @staticmethod
    def calculate_score_for_rule(
        rule: Rule, data_1: Dict[str, Any], data_2: Dict[str, Any]
    ) -> Optional[Dict[str, Any]]:
        """
        Calculate a matching score for one rule between FHIR Person-Person, or Person-Patient, or Person/Patient-AppUser
        :param rule: one rule in the generated rules by RulesGenerator
        :param data_1: Dictionary of Pii data for FHIR Person/Patient data, or AppUser data
        :param data_2: Dictionary of Pii data for FHIR Person/Patient data, or AppUser data
        :return: Dictionary of 1 rule score result
        """

        id_data_1: Optional[Any] = data_1.get(Attribute.ID)
        id_data_2: Optional[Any] = data_2.get(Attribute.ID)
        if not (id_data_1 and id_data_2):
            return None

        score_avg: float = 0.0
        for attribute in rule.attributes:
            val_1: Optional[Any] = data_1.get(attribute)
            val_2: Optional[Any] = data_2.get(attribute)

            if val_1 and val_2:
                # calculate exact string match on "trimmed lower" string values
                score_avg += fuzz.ratio(
                    str(val_1).strip().lower(), str(val_2).strip().lower()
                )

        score_avg /= len(rule.attributes)
        rule.score = score_avg

        rule_score_result: Dict[str, Any] = {
            Attribute.ID + "_1": str(id_data_1),
            Attribute.ID + "_2": str(id_data_2),
            Attribute.RULE_NAME: rule.name,
            Attribute.RULE_SCORE: rule.score,
        }

        return rule_score_result
